/*
 * An opener for a file provided by an Ext2-compatible filesystem.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <e2access/access.h>
#include <e2access/image.h>
#include <e2access/path.h>

#include <systypes/fcntl.h>

#include "ext2_file_accessor.h"
#include "ext2_file_opener.h"

/* Return a file identifier for the given 'path'. */

long Ext2FileOpener::get_fileid(const char *path, flags_t flags, fileid_t *fileid)
{
    const char *filename = path;

    /* Obtain the inode number. */

    ext2_ino_t ino;
    errcode_t retval = image_find_path(_fs, &filename, &ino);

    /* Handle a missing file. */

    if (retval)
    {
        /* Create a missing file if possible. */

        if (flags & O_CREAT)
        {
            /* Determine whether only the leafname is left of the path, with
               the inode number referring to the parent directory. */

            if (path_is_leafname(filename))
            {
                struct ext2_inode inode_parent;

                /* Determine write access in the directory. */

                retval = ext2fs_read_inode(_fs, ino, &inode_parent);

                if (retval)
                    return -L4_EIO;

                if (!access_can_write(_user, &inode_parent))
                    return -L4_EPERM;

                /* If so, use the parent directory inode to create a new file. */

                if (image_create_file(_fs, ino, filename, 0666 & ~_user.umask,
                                      _user.uid, _user.gid, &ino))
                    return -L4_EIO;

                *fileid = (fileid_t) ino;
                return L4_EOK;
            }
        }

        *fileid = FILEID_INVALID;
        return -L4_ENOENT;
    }

    *fileid = (fileid_t) ino;
    return L4_EOK;
}

/* Return a new accessor for 'fileid'. */

long Ext2FileOpener::make_accessor(const char *path, flags_t flags, fileid_t fileid, Accessor **accessor)
{
    (void) path; (void) flags;

    ext2_file_t file;
    errcode_t retval = ext2fs_file_open(_fs, (ext2_ino_t) fileid, EXT2_FILE_WRITE, &file);

    // NOTE: Map error conditions.

    if (retval)
        return -L4_EIO;

    *accessor = new Ext2FileAccessor(file, fileid);
    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
